# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

############################
# Standard cmdbar commands #
############################

# -------------- set config
proc :set {args} {
  if {[lindex $args 0] eq {aget}} {
    return [setup aget *]
  } elseif {[lindex $args 0] eq {aset}} {
    return [setup aset [lindex $args 1]]
  } elseif {[llength $args] == 1} {
    return [setup cget $args]
  } else {
    return [setup configure [lindex $args 0] [lindex $args 1]]
  }
}

# -------------- overlay management

proc :add {args} {
  set id [::overlay::add $args]
  return #$id
}

proc :modify {{ID {}} {args {}}} {
  if {$ID eq {}} {
    set id [lindex [::overlay::select get] 0]
  } elseif {[catch {isID $ID} id]} {
    return -code error $id
  }
  foreach {prop arg} $args {
    ::overlay::modify $id [list $prop $arg]
  }
}

proc :query {ID prop} {
  return [getProp $ID $prop]
}

proc :copy {{ID {}}} {
  if {$ID eq {}} {
    set id [lindex [::overlay::select get] 0]
  } elseif {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
    return -code error "Not a valid ID: $ID"
  }
  set sel [::overlay::copy $id]
  if {[llength $sel] == 1} {
    ::cmdbar::writeToCmdBar "Overlay #$sel copied to buffer" -bg green
  } elseif {[llength $sel] > 1} {
    ::cmdbar::writeToCmdBar "Copied [llength $sel] overlays to buffer" -bg green
  } else {
    ::cmdbar::writeToCmdBar {Nothing to copy} -bg green
  }
}

proc :paste {} {
  set IDs [list]
  foreach id [::overlay::paste] {
    set props [list Name "Copy of [getProp $id Name]"]
    ::overlay::modify $id $props
    lappend IDs #$id
  }
  # join the list with spaces, workaround because the first ID may have been added as a sublist
  return [join $IDs { }]
}

proc :rename {{ID {}} {new {}}} {
  if {$ID eq {}} {
    set id [lindex [::overlay::select get] 0]
  } elseif {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
    return -code error "Not a valid ID: $ID"
  }
  if {$new ne {}} {
    ::overlay::modify $id [list Name $new]
  } else {
    set win [::overlay::win]
    # ttk::treeview does not allow inplace editing
    # we place an entry just over the item
    lassign [$win bbox $id 0] x y width height
    set name [::ttk::entry $win.name -width $width]
    set b [::cmdbar::deleteBindings]
    grab set $name
    if {[catch {$name insert 0 [getProp $id Name]} msg]} {
      destroy $name
      ::cmdbar::restoreBindings $b
      focus $win
      return -code error $msg
    }

    $name selection range 0 end
    place $name -x $x -y $y
    focus $name

    bind $name <Return> [list apply {{win id name b} {
      ::overlay::modify $id [list Name [$name get]]
      destroy $name
      ::cmdbar::restoreBindings $b
      focus $win
    }} $win $id $name $b]

    bind $name <Escape> [list apply {{win name b} {
      destroy $name
      ::cmdbar::restoreBindings $b
      focus $win
    }} $win $name $b]
  }
  return -code ok $ID
}

proc :color {IDs {new {}}} {
  foreach ID $IDs {
    set id [isID $ID]
    if {$new eq {}} {
      if {[set color [tk_chooseColor -initialcolor [getProp $id Color] \
                    -title [getProp $id Name]]] eq {}} {
        # cancel              
        return -code ok $IDs
      } 
    } else {
      set color $new
    }
    ::overlay::modify $id [list Color $color]
  }
  return -code ok $IDs
}


proc :list {{flags all}} {
  set IDs [list]
  switch -- $flags {
    visible {
      foreach id [::overlay::ids] {
        if {[getProp $id Show] eq {yes}} {
          lappend IDs #$id
        }
      }
    }
    invisible {
      foreach id [::overlay::ids] {
        if {[getProp $id Show] eq {no}} {
          lappend IDs #$id
        }
      }
    }
    all {
      foreach id [::overlay::ids] {
        lappend IDs #$id
      }
    }
    default {
      set IDs [list]
    }
  }
  # join the list with spaces, workaround because the first ID may have been added as a sublist
  return [join $IDs { }]
}

# ------------------ Macro handling 
proc :macro {op args} {
  switch -- $op {
    load {
      foreach name $args {
        load_macro $name
      }
      return -code ok
    }
    exec {
      # save current config
      set conf [setup aget]
      set code [catch {::macro::exec [lindex $args 0] [lrange $args 1 end]} msg]
    }
    default {
      # interpret $op as the macro name
      set conf [setup aget]
      set code [catch {::macro::exec $op $args} msg]
    }
  }
  switch -- $code {
    0 {
      # everything's ok
    }
    1 {
      # play back config
      # in diff mode to not to trigger unneccessary traces
      setup aset $conf -diff
      # then raise the error
      return -code $code $msg
    }
    default {
      # any other return code:
      # play back the config and return gracefully
      setup aset $conf -diff
      return -code ok
    }
  }
}

# --------- window show/hide
proc :show {plot} {
  set win [winfo toplevel [::TkEsweepXYPlot::window $plot]]
  set x [expr {([winfo screenwidth .]-[winfo width $win])/2}]
  set y [expr {([winfo screenheight .]-[winfo height $win])/2}]
  wm geometry  $win +$x+$y
  #wm transient $win .
  updateAppTitle $plot
  # withdraw the window to make the window manager notice the change
  wm withdraw $win
  wm deiconify $win
  tkwait visibility [::TkEsweepXYPlot::window $plot]
  eval update$plot
}

proc :hide {plot} {
  wm withdraw [winfo toplevel [::TkEsweepXYPlot::window $plot]]
  update
}

proc :q {{code 0}} {
  ::config::writeSetup
  exit $code
}

proc :message {msg {type ok} {icon info} {title !}} {
  return [tk_messageBox -message $msg -type $type -icon $icon -title $title] 
}

proc :dialog {title prompt args} {
  #return [tk_dialog .dia_[clock microseconds] $title $text {} $default {*}$buttons]
  return [tk_dia -title $title -prompt $prompt {*}$args]
}

proc :getString {title prompt args} {
  return [tk_getString -title $title -prompt $prompt -default {} {*}$args]
}

# select a curve or returns the selected curve(s) id(s)
proc :select {{IDs {}}} {
  if {$IDs ne {}} {
    set op {set}
    foreach ID $IDs {
      puts stderr $ID
      if {$ID eq {end}} {
        ::overlay::select end
      } elseif {[catch {isID $ID} id]} {
        return -code error $id
      } else {
        ::overlay::select $op $id
      }
      # next IDs should be added
      set op {add}
    }
    set result [list]
    foreach id [::overlay::select get] {
      lappend result #$id
    }
    return $result
  } else {
    set IDs [list]
    foreach id [::overlay::select get] {
      lappend IDs #$id
    }
    return $IDs
  }
}

# returns the ID of the master curve
proc :master {{ID {}}} {
  if {$ID ne {}} {
    if {[catch {isID $ID} id]} {
      return -code error $id
    } else {
      return [set ::master $id]
    } 
  } else {
    if {$::master ne {}} {
      return #$::master
    } else {
      return -code error {Master curve not yet defined}
    }
  }
}

# simulate press of a single key. Multiple keys may be invoked in one command, like
# : key {press}
# where every keypress for "press" is generated
proc :key {pattern} {
  for {set i 0} {$i < [string length $pattern]} {incr i} {
    event generate . [string index $pattern $i]
  }
}

proc :pwd {} {
  return -code ok [pwd]
}

# repeat a command. T is the index to go back in history:
# T=0: last command
# T=1: last but on command
# T=2: third to last
# T=... and so on
#
# if T<0 the function returns the number of commands in history
proc :history {{T 0}} {
  if {$T < 0} {
    return -code ok [::cmdbar::historyGet $T]
  } else {
    ::cmdbar::addExecFlags [list -nodisable -nohistory] 
    ::cmdbar::enterCmdBar [::cmdbar::historyGet $T]
  }
}

# wait for an ID's property to be set
proc :wait {ID prop} {
  if {[catch {isID $ID} id]} {
    return -code error $id
  }
  vwait ::overlay::overlays($id,$prop)
}

proc :help {} {
  if {![file exists [set helpfile [file join [pwd] doc README.txt]]]} {
    return -code error {Help yourself!}
  }
  if {[winfo exists .help]} {
    wm deiconify .help
  } else {
    set h [toplevel .help]
    wm withdraw $h
    wm transient $h .
    set txt [text ${h}.txt]
    pack $txt -expand yes -fill both
    set fp [open $helpfile r]
    $txt insert end [read $fp]
    close $fp
    wm deiconify $h
  }
  return -code ok
}

# general helper functions
proc getProp {ID prop {subIdx 1}} {
  if {[catch {isID $ID} id]} {
    if {![regexp {^[[:digit:]]+$} $ID id]} {
      # compatibility for old versions
      return -code error "Not a valid ID or number: $ID"
    }
  }
  return [lindex [::overlay::get $id $prop] end $subIdx]
}

proc isID {ID} {
  if {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
    return -code error "Not a valid ID: $ID"
  } else {
    return $id
  }
}

proc load_macro {{name {}}} {
  if {$name eq {}} {
    foreach fname [glob -nocomplain -directory [file join [pwd] macros] *.frm] {
      ::macro::import $fname
    }
  } else {
    set fname [file join [pwd] macros $name.frm]
    ::macro::import $fname
  }
}

