# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide spice 1.0

lappend auto_path [file join [pwd] ..]
package require complex
package require cmatrix

namespace eval ::spice {
}

# parse the netlist into a format which is easily converted into a matrix
# the netlist is a typical spice netlist, format
# R1  1 2   1
# C1  2 0   1u
# This is a simple RC-Lowpass of a resistor with 1 Ohms and a capacitor with 1 uF
proc ::spice::parseNetlist {netlist} {
  # Step 1: reorder the netlist to have all parts connected to a node in a single list
  set i 0
  set nodes [list]
  foreach line $netlist {
    lassign $line name n1 n2
    if {$n1 == $n2} {
      return -code error "Syntax error in line $i: component connects to itself"
    }
    set params [lrange $line 3 end]
    # we can now convert the part type to a model function
    # the type is the the first part of the name up to the first number
    if {[regexp -- {^[[:alpha:]]+} $name type] == 0} {
      return -code error "Syntax error in line $i: type definition not parseable"
    }
    switch -- $type {
      R {
        # function for real part
        set rfunc "1/[prefix2double [lindex $params 0]]"
        # function for imaginary part
        set ifunc 0.0
      }
      C {
        set rfunc 0.0
        set ifunc "\$w*[prefix2double [lindex $params 0]]"
      }
      L {
        set rfunc 0.0
        set ifunc "1/(\$w*[prefix2double [lindex $params 0]])"
      }
      default {
        continue
      }
    }
    # append this functions to a node, and remember the node on the other side of the component
    lappend nodes_$n1 [list $n2 [list $rfunc $ifunc]]
    if {[lsearch $nodes $n1] < 0} {
      lappend nodes $n1
    }
    incr i
  }
  # we need to remap the nodes; it will be easier to build the matrix
  # replace each node with the index of its occurence in the nodes list
  foreach n $nodes { 
    set leg [list]
    foreach comp [set nodes_$n] {
      if {[set i [lsearch -integer $nodes [lindex $comp 0]]] < 0} {
        set i 0
      } else {
        incr i
      }
      lset comp 0 $i
      lappend leg $comp
    }
    set nodes_$n $leg
  }

  #set nodes [lsort -increasing -integer $nodes]
  set i 1
  foreach n $nodes {
    lappend netdef [list $i [set nodes_$n]]
    incr i
  }

  return $netdef
}

# build a matrix out of the previously parsed netlist
proc ::spice::buildMatrix {netdef w} {
  # create mxm matrix
  puts $netdef
  ::cmatrix::create M [set m [llength $netdef]] $m
  ::complex::cset neg -1
  # get one component after the other, evaluate at circular frequency w, and insert into matrix
  for {set i 0} {$i < $m} {incr i} {
    set leg [lindex $netdef $i]
    set row [lindex $leg 0]
    incr row -1
    # in <row,row> is the sum of all conductances
    # in <row,n> is the negative conductance between node $row and node $n
    set Z [::cmatrix::mget $M $row $row]
    foreach comp {*}[lrange $leg 1 end] {
      lassign $comp col X
      incr col -1
      # evaluate X at circular frequency w
      lset X 0 [expr [lindex $X 0]]
      lset X 1 [expr [lindex $X 1]]
      set Z [::complex::cadd $Z $X]
      # components to ground are only summed
      if {$col < 0} {continue}
      set Y [::complex::cmul $neg [::complex::cadd [::cmatrix::mget $M $row $col] $X]]
      ::cmatrix::mset M $row $col $Y
    }
    ::cmatrix::mset M $row $row $Z
  }
  ::cmatrix::toStr $M

  return $netdef
}


# convert a number appended with a unit prefix to a double number
# Example: prefix2double 1u => 1e-6
proc prefix2double {number} {
  if {[regexp -- {^([[:digit:]\.]+)([[:alpha:]]+)$} $number -> num prefix] == 0} {
    # it's already a number
    return $number
  }
  switch -- $prefix {
    a {set fact 1e-18}
    f {set fact 1e-15}
    p {set fact 1e-12}
    n {set fact 1e-9}
    u {set fact 1e-6}
    m {set fact 1e-3}
    c {set fact 1e-2}
    d {set fact 1e-1}
    da {set fact 1e+1}
    h {set fact 1e+2}
    k {set fact 1e+3}
    M {set fact 1e+6}
    G {set fact 1e+9}
    T {set fact 1e+12}
    P {set fact 1e+15}
    E {set fact 1e+18}
    default {
      return -code error "Cannot resolve unit prefix of $number"
    }
  }
  if {[catch {set num [expr {$num*$fact}]} err]} {
    return -code $err
  } else {
    return $num
  }
}

proc test {} {
  set netlist [list [list R1 1 21 1] [list C1 21 0 1u]]
  set netdef [::spice::parseNetlist $netlist]
  ::spice::buildMatrix $netdef 10
}

test

